#ifndef BOARD_INCLUDED
#define BOARD_INCLUDED

#include "position.hh"
#include "server_image.hh"
#include "square.hh"
#include "inline.hh"
#include "scaled_board_image.hh"
#include "context_menu.hh"
#include "board_settings_editor.hh"
#include "owned_editor.hh"
#include "board_settings_stream.hh"
#include "persistent.hh"

class Board
	:  virtual public Widget,
	   virtual public Menued
{
public:

	enum {menu_main};
  
	class Open_Info
	{
	public:

		Pointer <const Position> position;
		int msg_move;
		int msg_middle;
		int msg_turn;
	
		Open_Info(const Position &new_position,
			  int new_msg_move= -1,
			  int new_msg_middle= -1,
			  int new_msg_turn= -1)
			:  position(new_position),
			   msg_move(new_msg_move),
			   msg_middle(new_msg_middle),
			   msg_turn(new_msg_turn)
			{
			}
	};

	Board();
	~Board()
		{
		}

	void Menued_Get(int menu, Menued::Entries &entries);

	bool Open(Pointer <Widget> parent, 
		  const Open_Info &);
	/* MSG_MOVE is sent after a piece was dragged.  The squares
	   can be retrieved with Get_Squares().  A message is sent
	   even when the piece was released outside the board.
	   In this case Get_Squares() returns DRAG_TO=0. 

	   After a move is played the two squares involved must be updated. 
	   
	   When the middle mouse button is clicked, MSG_MIDDLE is sent
	   and Get_Squares() returns the square in FROM. 
	*/

	void Invalidate_Square(int square,
			       bool interrupt_dragging= true);
	void Invalidate_Position();
	/* All of POSITION has changed. 
	 */

	bool Get_Squares(int &from, int &to);
	/* After MSG_MOVE, this functions tells the squares.
	   Return FALSE if no piece was dragged.  
	*/

	const Board_Settings &Get_Settings() const
		{
			return persistent_settings.Get_Value();
		}
	void Set_Turned(bool new_turned); 
	bool Get_Turned() const
		{
			return persistent_settings().flags & bs_turned;
//			return settings.turned;
		}

	void Set_Selection(int square_a, int square_b);
	void Clear_Selection();
	void Invalidate_Move_Squares(const Move &);

private:

	enum {msg_square_names= Widget::msg_custom,
	      msg_turn_board,
//	      msg_edit_settings,
//	      msg_settings_edited,
	      msg_anti_alias,
	      msg_frame,
	      msg_settings,

	      msg_custom};

	int square_width;
	Scaled_Board_Image scaled_image;
	Server_Image board_image;
	/* The whole board, including the frame if it is visible. 
	 */
	Context_Menu menu;
	Pointer <const Position> position;
	int msg_move;
	int msg_middle;
	int msg_turn;
	bool dragging;
	Point drag_point;
	Point pointer_pos;
	/* Position of pointer inside the dragged square. 
	 */
	Point last_org;
	/* Origin of dragged piece rectangle. 
	 */
	int drag_from;
	int drag_to;
	Piece dragged_piece;
	Server_Image drag_img;
	/* Same size as BOARD_IMAGE.  Background image when dragging. 
	 */
	Persistent <Board_Settings, Board_Settings_Editor> persistent_settings;
//	Owned_Editor <Board_Settings_Editor> edit;
//	Board_Settings settings;
	int selection[2];
	bool have_colors; 
	Canvas::Color colors[3];
	/* Index as in Settings::colors.  Set if HAVE_COLORS.  Not
	   used for frame color. 
	 */
//	bool background_invalid;

	static int frame_width;

	void Change_Sized_Image();
	/* BOARD_IMAGE must be.  Set SQUARE_WIDTH to -1 before calling to
	   force new loading. 
	 */
	void Draw_All();
	void Draw_Square(int file, 
			 int rank, 
			 bool always_empty= false);
	void Draw_Square_Background(Canvas &, 
				    int file, 
				    int rank, 
				    Rect rect);
	/* RECT must correspond to the square on the board.  
	 */

	void Draw_Frame();

	int Point_Square(Point) const;
	Rect   Square_Rect(int file, int rank) const;
	void Interrupt_Dragging();

	void MR_Receive(int);
	void Wdg_Draw(Canvas &);
	void Wdg_Size(Point);
	void Wdg_Mouse(Mouse_Action, Mouse_Button, Point, bool);
	void Wdg_Mouse_Cursor();
	void Wdg_Close();
	void Wdg_Get_Size_Request(int &, int &, bool &); 

	Canvas::Color Get_Square_Color(Chess_Color cc) const
		{
			Assert (cc >= 0 && cc < 2);
			Assert (have_colors);
			return colors[cc];
		}
	bool Is_Inside_Point(Point p) const;
};

#endif /* ! BOARD_INCLUDED */
