#include "evaluator.hh"

/* Needs:  chess_evaluator    */

#include "fen.hh"
#include "parse_move.hh"
#include "parser.hh"
#include "error.hh"
#include "get_dir.hh"
#include "file_name.hh"
#include "score.hh"

bool
Evaluator::Run()
{
	Assert (! IO_Child_Process::Is());

	if (has_started)
	{
		return true;
	}

	String fen= FEN(position());

	String program_name;
	Vector <String> args;

	if (settings.engine.Is())
	{
		program_name= Get_Dir(id_bin) / "baby_xboard_chess_evaluator"; 
		args += "-e";
		args += settings.engine; 
	}
	else
	{
		program_name= Get_Dir(id_bin) / "baby_chess_evaluator"; 
	}

	args += fen;

	if (! IO_Child_Process::Open(program_name, 
				     args,
				     settings.nice))
		return false;

	has_started= true;
  
	EC_Run();

	output_reader.Open(IO_Child_Process::Get_Output());
	error_output_reader.Open(IO_Child_Process::Get_Error_Output());

	generator.New(position());
	best_move= -1;
	Timer::Start(milliseconds);
	start_date= Get_Date();
	knots= 0;
	depth= 0;
	score= 0;
	last_time_ms= 0;

	return true;
}

void
Evaluator::Pause()
{
	if (IO_Child_Process::Is())
	{
		IO_Child_Process::Pause();

		if (IO_Child_Process::Paused())
		{
			if (Timer::Is())
				Timer::Stop();
			age= Date_Diff_ms(start_date, Get_Date());
		}
		else
		{
			start_date= Get_Date() + ms_To_Date_Diff(- age);
			Timer::Start(milliseconds);
		}
		EC_Pause(); 
	}
}

void
Evaluator::IOCP_Output(String line)
{
	Parser parser(line); 

	parser.Skip_Space(); 

	String word= parser.Skip_Word(); 

	if (word == "depth")
	{
		parser.Skip_Space();
		String value= parser.Skip_Word(); 
		if (value.Is())
		{
			int new_depth;
			if (1 == sscanf(value(), "%d", & new_depth))
			{
				if (new_depth > depth)
				{
					depth= new_depth;
					EC_Depth(depth);
				}
			}
		}
	}
	else if (word == "main")
	{
		parser.Skip_Space();
		String move= parser.Skip_Word(); 
		int new_move= Parse_Move(move(), generator);
		if (new_move != -1)
		{
			best_move= new_move;
			EC_Main(new_move);
		}
	}
	else if (word == "score")
	{
		parser.Skip_Space();
		String new_score_string= parser.Skip_Word(); 

		score= Parse_Score(new_score_string);

		EC_Score(score);
	}
	else if (word == "knots")
	{
		parser.Skip_Space();
		String value= parser.Skip_Word();

		int new_knots;
		if (1 == sscanf(value(),
			      "%d", 
			      & new_knots))
		{
			knots= new_knots;
		}

		EC_Knots(knots);
	}
}

void
Evaluator::IOCP_Error_Output(String)
{
}

bool
Evaluator::Re_Start()
{
	Terminate(); 
	return Run();
}

void 
Evaluator::Timer_Alarm()
{
	Dbg_Function();

	Date now= Get_Date();
	int date_ms= Date_Diff_ms(start_date, now);

	Process_Usage usage(Get_IO_Child().Get_Process_Usage());
	int time_ms= usage.total + usage.children_total; 
	Dbg_Int(time_ms);
	last_time_ms >?= time_ms;
	
	EC_Timer(date_ms, last_time_ms);

	/* Terminate if necessary. 
	 */
	if (settings.timing == Evaluator_Settings::ti_time   &&
	    date_ms >= 1000 * settings.time_s)
	{
		Terminate(false, true);
	}
}

void
Evaluator::Terminate(bool really,
		     bool engine_finished)
{
	Dbg_Function();
	Dbg("\treally= %d", really);

	if (IO_Child_Process::Is())
	{
		output_reader.Close();
		error_output_reader.Close();

		if (Timer::Is())
			Timer::Stop();
		IO_Child_Process::Close();
		EC_Terminate(really, engine_finished); 
		if (really)
			has_started= false;
	}

	else if (really && has_started)
	{
		has_started= false;
		EC_Terminate(really, engine_finished);
	}
}

void
Evaluator::Input()
{
	if (Proceed())
	{
		Terminate(false, true);
	}
}

Evaluator::Evaluator()
	:  output_reader(this),
	   error_output_reader(this)
{
}



