#ifndef POSITION_INCLUDED
#define POSITION_INCLUDED

/* Chess position. 

   The position is not necessarily legal, but consistent:  MATERIAL
   corresponds to PIECES etc...
 */

#include "square.hh"
#include "piece.hh"
#include "move.hh"
#include "pointer.hh"
#include "material.hh"

class Position
	:  public Pointed
{
public:

	Position() 
		/* clear */
		{
			Init_Outside(); 
			to_move= 0;
			moved[1]= 0;
			moved[4]= 0;
			Clear();
		}
	Position(int /* dummy */ ) 
		/* From Initial position */
		{
			Init_Outside();  
			Reset(); 
		}
	~Position() 
		{
		}

	Piece &Get_Piece(int square)
		{
			Assert (square >= 0 && square < 120);
			return pieces[square];
		}
	Piece Get_Piece(int square) const
		{
			Assert (square >= 0 && square < 120);
			return pieces[square];
		}
	Piece &operator [] (int square)
		{
			return Get_Piece(square); 
 		}
	Piece operator [] (int square) const
		{
			return Get_Piece(square); 
		}
	Piece *Get_Pieces() 
		{
			return pieces; 
		}
	const Piece *Get_Pieces() const 
		{
			return pieces; 
		}
	int &Get_Moved(int index)
		{
			Assert(0 <= index && index < 6);
			return moved[index];
		}
	int Get_Moved(int index) const
		{
			Assert(0 <= index && index < 6);
			return moved[index];
		} 
	int &Get_King_Square(Chess_Color cc)
		{
			Assert(cc == cc_w || cc == cc_b);
			return king_squares[cc];
		}
	int Get_King_Square(Chess_Color cc) const
		{
			Assert (cc == cc_w || cc == cc_b);
			return king_squares[cc];
		}

	void Reset(); 
	/* Reset to the initial position. */

	void Clear();
	/* Don't change MOVED. 
	 */

	void operator << (const Move &); 
	/* Play a move      */
	void operator >> (const Move &); 
	/* Take back a move */

	bool Square_Attacked(int square, 
			     Chess_Color color)
		const;
	/* Whether COLOR is attacked on SQUARE.
	 */

	bool Is_Check() const
		{
			return Square_Attacked(Get_King_Square(to_move), to_move);
		}
	/* Whether the king of the side to move is in check. 
	 */

	void Update_Material();
	void Remove_Piece(int square)
		{
			Assert (Is_Inside_Square(square));
			Piece piece= pieces[square];
			-- material(piece.Get_Type(), piece.Get_Color());
			pieces[square]= Piece(false);
		}
	Chess_Color Get_To_Move() const
		{
			return to_move; 
		}
	void Transfer_Piece(int from, int to);
	void Change_To_Move(); 
	int Get_EP_Square() const
		{
			return ep_square; 
		}
	int &Get_EP_Square()
		{
			return ep_square; 
		}
	void Insert_Piece(int square, Piece piece); 
	int Get_Previous_Move_Count() const
		{
			Assert ((previous_move_count + to_move) % 2 == 0); 
			return previous_move_count;
		}
	void Set_Previous_Move_Count(int new_previous_move_count)
		{
			Assert (new_previous_move_count >= 0);
			Assert ((new_previous_move_count + to_move) % 2 == 0); 
			previous_move_count= new_previous_move_count;
		}
	int Get_Halfmove_Clock() const
		{
			return halfmove_clock; 
		}
	int Get_Material(int pt, int cc) const
		{
			return material(pt, cc); 
		}
	Material &Get_Material()
	  {
	    return material; 
	  }
	void Set_To_Move(int new_to_move)
		{
			if (new_to_move == to_move)
				return;
			Change_To_Move(); 
		}
	const Material &Get_Material() const
		{
			return material; 
		}
	void Set_Halfmove_Clock(int new_halfmove_clock)
		{
			halfmove_clock= new_halfmove_clock; 
		}
	
private:

	Chess_Color to_move;
	/* who is to move */

	int ep_square;
	/* Where the side to move can capture "en passant"; for instance
	   e5.  0 if there is no "en passant" move.
	*/

	Piece pieces[120];
	/* The index is the square number, as defined in SQUARE.
	   Squares outside the board are initilized by the constructor
	   and are never changed, but can be read. 
	 */

	int moved[6]; 
	/* By index:  a1, e1, h1, a8, e8, h8 

	   For each of these squares, the number of times a piece has
	   moved on it.  If 0, the square contains the rook or king
	   from the initial position.   
	   
	   When editing a position, moved[e1] and moved[e8] should
	   always be 0, so you have:
	   wO-O   <=> moved[0] == 0
	   wO-O-O <=> moved[2] == 0
	   bO-O   <=> moved[3] == 0
	   bO-O-O <=> moved[5] == 0
 	*/

	int king_squares[2];
	/* Position of kings. 
	 */

	int previous_move_count; 
	/* Number of halfmoves since beginning of game.  Even if white
	   is to move, else odd. 
	 */

	int halfmove_clock; 

	Material material;
	/* Piece count.  In sync with PIECES. 
	 */

	void Init_Outside();
	/* Set the square outside the chess board to the appropriate
	   value.  Called once in the constructor. 
	 */ 
};

#endif /* ! POSITION_INCLUDED */
