#include "string.hh"

/* A string is either one of:

   - Empty:      raw == NULL
   - Constant:   Constant(raw)
   - Owned:      ! Constant(raw)

   If the string is constant, the place RAW points to is supposed to
   never change.  

   If the string is owned, HEAD points to a chunk of of memory that
   can hold HEAD->PLACE characters.  HEADs are shared between several
   String objects using reference counting. 
 */

#include "deref.hh"

void
String::Free_Head(Head *head)
{
	Assert (head != NULL && ! Constant(head));
	Assert (Deref(head).ref_count >= 1);

	if (Deref(head).ref_count > 1)
	{
		-- head->ref_count;
	}
	else
	{
		Assert (Deref(head).ref_count == 1);
		Mem::Free(head);
	}
}

bool
String::Do_Assign(const char *string)
{
	Assert(string != NULL);

	int length= strlen(string);

	if (Constant(string))
	{
		constant= string;
	}
	else
	{
		head= (Head *)Mem::Alloc
			(sizeof(Head) + (1 + length) * sizeof(char));
		if (head == NULL)
			return false;

		Deref(head).length= length;
		Deref(head).ref_count= 1;
		Deref(head).place= length;
		memcpy(Deref(head).string,
		       string,
		       (1 + length) * sizeof(char));
	}
	return true;
}

String::String(const String &string)
	:  Pointed()
{
	if (! string.Is())
	{
		raw= NULL;
		return;
	}

	if (Constant(string.raw))
	{
		constant= string.constant;
	}
	else
	{
		head= string.head;
		++ Deref(head).ref_count;
	} 
}

String::String(const char *string)
{
	if (string != NULL &&
	    string[0] != '\0')
	{
		if (Constant(string))
		{
			constant= string;
		}
		else
		{
			int length= strlen(string);
			Assert (length >= 1);
			head= (Head *)Mem::Alloc
				(sizeof(Head) + (length + 1) * sizeof(char));
			if (head != NULL)
			{

				Deref(head).length= length;
				Deref(head).ref_count= 1;
				Deref(head).place= length;
				memcpy(Deref(head).string, 
				       string, 
				       (length + 1) * sizeof(char));
			}
		}
	} 
	else
		raw= NULL;
}

bool
String::Isolate()
{
	if (raw == NULL)
		return true;

	if (Constant(raw))
	{
		int length= strlen(constant);
		Head *new_head= (Head *)Mem::Alloc
			(sizeof (Head) + (1 + length) * sizeof (char));
		if (new_head == NULL)
			return false;
		Deref(new_head).length= length;
		Deref(new_head).place= length;
		Deref(new_head).ref_count= 1;
		memcpy(Deref(new_head).string,
		       constant,
		       (1 + length) * sizeof (char));
		head= new_head;
		return true;
	}
	else
	{
		if (Deref(head).ref_count == 1)
			return true;

		Head *new_head= (Head *)Mem::Alloc
			(sizeof(Head) + (1 + Deref(head).length) * sizeof (char));

		if (new_head == NULL)
			return false;

		Deref(new_head).length= Deref(head).length;
		Deref(new_head).ref_count= 1;
		Deref(new_head).place= Deref(head).length;
		memcpy(Deref(new_head).string,
		       Deref(head).string,
		       (1 + Deref(head).length) * sizeof (char));
		-- Deref(head).ref_count;
		head= new_head;
	
		return true;
	}
}

const String &
String::operator = (const String &string)
{
	if (this == &string)
		return This;

	if (raw != NULL && ! Constant(raw))
		Free_Head(head);

	if (! string.Is())
	{
		raw= NULL;
		return This;
	}

	if (Constant(string.raw))
	{
		constant= string.constant;
	}
	else
	{
		head= string.head;
		++ Deref(head).ref_count;
	} 

	return This;
}

const String &
String::operator= (const char *string)
{
	if (string != NULL && string[0] == '\0')
		string= NULL;
	if (raw != NULL && ! Constant(raw))
		Free_Head(head);
	if (string == NULL)
		raw= NULL;
	else
		Do_Assign(string);

	return This;
}

String 
String::operator + (const String &string) 
	const
{
	if (raw == NULL)
		return string;

	if (string.raw == NULL)
		return This;

	int own_length= *This;
	int length= *string;

	String result(own_length + length, 0);
	
	if (*result != own_length + length)
		return result;

	memcpy(result.Get_Ref(),
	       This(),
	       own_length);
	memcpy(result.Get_Ref() + own_length,
	       string(),
	       length);
	return result;
}

String 
String::operator + (const char *string) const
{
	if (string != NULL && string[0] == '\0')
		string= NULL;

	if (raw == NULL)
		return string;

	if (string == NULL)
		return This;

	int own_length= *This;
	int length= strlen(string);

	String result(own_length + length, 0);
	
	if (*result != own_length + length)
		return result;

	memcpy(result.Get_Ref(),
	       This(),
	       own_length);
	memcpy(result.Get_Ref() + own_length,
	       string,
	       length);
	return result;
}

String &
String::operator += (const char *string)
{
	if (raw == NULL)
	{
		if (Constant(string))
		{
			constant= string;
		}
		else
		{
			int length= strlen(string);
			head= (Head *)Mem::Alloc
				(sizeof (Head) + (1 + length) * sizeof (char));
			if (head == NULL)
				return This;
			Deref(head).length= length;
			Deref(head).ref_count= 1;
			Deref(head).place= length;
			memcpy(Deref(head).string,
			       string,
			       (1 + length) * sizeof (char));
		}
	}
	else if (Constant(raw))
	{
		int own_length= strlen(constant);
		int length= strlen(string);
		Head *new_head= (Head *)Mem::Alloc
			(sizeof (Head) + (1 + own_length + length) * sizeof (char));
		if (new_head == NULL)
			return This;
		Deref(new_head).length= own_length + length;
		Deref(new_head).ref_count= 1;
		Deref(new_head).place= own_length + length;
		memcpy(Deref(new_head).string,
		       constant,
		       own_length * sizeof (char));
		memcpy(Deref(new_head).string + own_length,
		       string,
		       1 + length);
		head= new_head;
	}
	else 
	{
		if (! Isolate())
			return This;
		int own_length= Deref(head).length;
		int length= strlen(string);
		if (own_length + length <= Deref(head).place)
		{
			memcpy(Deref(head).string + own_length,
			       string,
			       (1 + length) * sizeof (char));
			Deref(head).length= own_length + length;
		}
		else
		{
			int new_place= Deref(head).place * 21 / 20 >? own_length + length;
			Head *new_head= (Head *)Mem::Re_Alloc
				(head,
				 sizeof (Head) + (1 + new_place) * sizeof (char));
			if (new_head == NULL)
				return This;
			head= new_head;
			Deref(head).length= own_length + length;
			Deref(head).place= new_place;
			memcpy(Deref(head).string + own_length,
			       string,
			       (1 + length) * sizeof (char));
		}
	}

	return This;
}

bool
String::operator == (const String &string) const
{
	return (raw == NULL
		? string.raw == NULL
		: (string.raw == NULL
		   ? false
		   : 0 == strcmp(Constant(raw) ? constant : Deref(head).string, 
				 Constant(string.raw)
				 ? string.constant 
				 : Deref(string.head).string))); 
}

String::String(int size, 
	       int /* dummy */)
{
	head= (Head *)Mem::Alloc(sizeof (Head) + (size + 1) * sizeof (char));
	if (head == NULL)
		return;
	Deref(head).length= Deref(head).place= size;
	Deref(head).ref_count= 1;
	Deref(head).string[size]= '\0';
}

String::String(const char *string, 
	       int size)
{
	Assert (size >= 0);
	
	if (size == 0)
	{
		raw= NULL;
		return;
	}

	head= (Head *)Mem::Alloc(sizeof (Head) + (size + 1) * sizeof (char));
	if (head == NULL)
		return;
	Deref(head).length= Deref(head).place= size;
	Deref(head).ref_count= 1;
	memcpy(Deref(head).string, string, size);
	Deref(head).string[size]= '\0'; 
}

String::String(char c)
{
	head= (Head *)Mem::Alloc(sizeof (Head) + 2 * sizeof (char));
	if (head == NULL)
		return;
	Deref(head).length= Deref(head).place= 1;
	Deref(head).ref_count= 1;
	Deref(head).string[0]= c;
	Deref(head).string[1]= '\0';
}

String &
String::operator += (const String &string)
{
	if (string.raw == NULL)
		return This;

	if (raw == NULL)
		This= string;
	else if (Constant(raw))
	{
		int own_length= strlen(constant);
		int length= *string;
		Head *new_head= (Head *)Mem::Alloc
			(sizeof (Head) + (1 + own_length + length) * sizeof (char));
		if (new_head == NULL)
			return This;
		Deref(new_head).length= Deref(new_head).place= length + own_length;
		Deref(new_head).ref_count= 1;
		memcpy(Deref(new_head).string,
		       constant,
		       own_length);
		memcpy(Deref(new_head).string + own_length,
		       Constant(string.raw) ? string.constant : Deref(string.head).string,
		       length + 1);
		head= new_head;
			 
	}
	else
	{
		if (Deref(head).ref_count == 1)
		{
			int own_length= Deref(head).length;
			int own_place= Deref(head).place;
			int length= string.Get_Length();
			if (length + own_length <= own_place)
			{
				memcpy(Deref(head).string + own_length,
				       string(),
				       length + 1);
				Deref(head).string[own_length + length + 1]= '\0'; 
				Deref(head).length += length;
			}
			else
			{
				int new_place= (own_length + length) >? own_length * 21 / 20;
				Head *new_head= (Head *)Mem::Re_Alloc
					(head,
					 sizeof (Head) + (1 + new_place) * sizeof (char));
				if (new_head == NULL)
					return This;
				head= new_head;
				memcpy(Deref(head).string + own_length,
				       string(),
				       length + 1);
				Deref(head).place= new_place;
				Deref(head).length= own_length + length;
			}
		}
		else
		{
			int own_length= *This;
			Head *new_head= (Head *)Mem::Alloc
				(sizeof (Head) + (1 + own_length + *string) * sizeof (char));
			if (new_head == NULL)
				return This;
			Deref(new_head).ref_count= 1;
			Deref(new_head).length= Deref(new_head).place= own_length + *string;
			memcpy(Deref(new_head).string,
			       Deref(head).string,
			       own_length);
			memcpy(Deref(new_head).string + own_length,
			       string(),
			       *string + 1);
			-- Deref(head).ref_count;
			head= new_head;
		}
	}
	return This;
}

void String::Free(void *raw)
{
	if (raw == NULL)
		return;
	if (Constant(raw))
		return;

	Free_Head((Head *) raw);
}

String String::Sub(int begin, 
		   int length) 
	const
{
	Assert (length >= 0);
	Assert (begin >= 0);
	Assert (begin + length <= Get_Length());

	if (length == 0)
		return "";

	String result(length, 0);

	memcpy(result.Get_Ref(),
	       This() + begin,
	       length);

	return result;
}
